library IEEE;
use IEEE.std_logic_1164.ALL;
use IEEE.numeric_std.ALL;

-- uMark, an 8008 computer with a Mark-8 control panel
--
-- Hardware:
--  i8008 or i8008-1
--  8KB RAM, 8KB Flash ROM
--  Xilinx XC2S15 Spartan-II FPGA (system controller)
--
-- Design:
--
--  The main difference between this computer and other 8008-based systems is
--  that it uses modern technology around the 8008.  High density RAM and in-
--  circuit-programmable Flash ROM completely fill the memory map, and an FPGA
--  performs almost all the functions that MSI logic would have in an earlier
--  design.
--
--  The FPGA has three major functions:
--   1) It implements all the control, latching and glue logic that the 8008
--      requires for memory, input/output and interrupt handling.
--   2) It manages an implementation of the Mark-8 front panel.  The
--      front panel uses momentary pushbutton switches with feedback LEDs
--      instead of paddle switches.  This is to provide for PC control of the
--      front panel, and to save space and cost of the board.
--   3) The FPGA provides an interface to a PC so that the PC can load
--      and store programs, and even act as an ICE debugger (TBD).
--
--  The 8008's bus is connected only to the RAM, ROM and FPGA, so no buffering
--  is necessary.  However, since Voh-min of the FPGA is 2.4V and Vih of the 8008
--  is 3.5V, the FPGA data bus outputs are open collector with a pull-up to 5V.
--  The ROM and RAM are 5V devices and meet the 8008's requirements.
--

entity logic is
    Port ( Clk : in std_logic;

           DBcpu : inout std_logic_vector(7 downto 0);
           Scpu : in std_logic_vector(2 downto 0);
           SYcpu : in std_logic;
           RDYcpu : out std_logic;
           INTcpu : out std_logic;

           ph1CPU, ph2CPU: out std_logic;

           AddrB : out std_logic_vector(15 downto 0);
           CERAM_n : out std_logic;
           CEROM_n : out std_logic;
           we_n: out std_logic;

           LED_c: out std_logic_vector(7 downto 0);
           LED_a: out std_logic_vector(5 downto 0);

           io_pio: inout std_logic_vector(7 downto 0);
           io_strobe: out std_logic_vector(3 downto 0);

           sw1in: in std_logic_vector(7 downto 0);
           sw2in: in std_logic_vector(7 downto 0);
           
           pc_strobe: in std_logic;
           pc_write: in std_logic;
           pc_addr: in unsigned(5 downto 0);
           pc_data: out std_logic);
end logic;

architecture RTL of logic is

    component DPRegister is
        Port ( clka, cea, da, clkb, ceb, db : in std_logic;
               q : out std_logic);
    end component;
    
    component Debouncer is
        Port ( clk, key : in std_logic;
               q : out std_logic);
    end component;

    component FD is
    Port ( C, CE, D : in std_logic;
           Q : out std_logic);
    end component;

    component FDC is
        Port ( C, CE, D, CLR : in std_logic;
               Q : out std_logic);
    end component;

    component FDCP is
        Port ( C, CE, PRE, CLR, D : in std_logic;
               Q : out std_logic);
    end component;

-- PC interface

    --  0.. 7   dl
    --  8..15   dh
    -- 16..23   switch register
    -- 24..31   control switches
    -- 56..63   data bus (examine)
    
    signal reg_load: std_logic_vector(31 downto 0);
    signal reg_data: std_logic_vector(31 downto 0);

-- processor management

    signal clkdiv: unsigned(3 downto 0);
    signal ph1, ph2: std_logic;
    
    signal st1i, st1, st2, st3, st4, st5, ststop, stwait: std_logic;
    signal st3_n, cc1, cc2, lal, lah: std_logic;
    
    signal dh: unsigned(7 downto 0);
    signal dl: unsigned(7 downto 0);
    
    signal stag, t3a, pcw, pcc, mem, ijam, syncint: std_logic;
    signal st: boolean;

    signal intr, step1: std_logic;
    signal DBmpx: std_logic_vector(7 downto 0);

    subtype FullAddress is std_logic_vector(15 downto 0);
    signal RAMbank, ROMbank: std_logic_vector(2 downto 0);

-- I/O decoding

    signal io: boolean;
    signal portsel: std_logic_vector(2 downto 0);
    signal port0: std_logic_vector(7 downto 0);

-- UI

    signal scanclk: unsigned(11 downto 0);
    signal uiclk, dhl_incr: std_logic;

    type UIState is (LEDhi, LEDlo, LEDo, LEDp0, LEDsw1, LEDsw2, SW1, SW2);
    signal state: UIState;
    signal LEDstrobes: std_logic_vector(5 downto 0);

    type DEPEXState is (DEStop, DELow, DELatchLow, DEHigh, DELatchHigh);
    signal DEPEX: DEPEXState;
    signal addrg: unsigned(8 downto 0);
    signal carry: unsigned(0 downto 0);
    signal DEPwrite, DEPflag1, DEPflag2, EXflag1, EXflag2: std_logic;
    signal latchdl, latchdh: std_logic;
    
-- switch inputs

    signal datasw, ctlsw, ctlsw_in: std_logic_vector(7 downto 0);
    signal tgl_datasw, tgl_ctlsw: std_logic_vector(7 downto 0);

    signal lahsw, lalsw, depsw, exsw, runstepsw, stepsw, intsw, jamintsw: std_logic;

-- switch debouncing

    signal sw1clk, sw2clk: std_logic;

-- functions so RAM/ROM mapping can be easily changed

    impure function RAMSelect return std_logic is
    begin
        if dh(5) = '0' then
            return '1';
        end if;
        return '0';
    end function;

    impure function RAMAddress return FullAddress is
    begin
        return (RAMbank & std_logic_vector(dh(4 downto 0) & dl(7 downto 0)));
    end function;

    impure function ROMAddress return FullAddress is
    begin
        return (ROMbank & std_logic_vector(dh(4 downto 0) & dl(7 downto 0)));
    end function;

begin

-- dual port register file - port A for control logic use, port B for PC interface

dlreg: for r in 0 to 7 generate
begin
    dlbit: component DPRegister port map (da => DBmpx(r), clka => clk, cea => lal,
                                          db => pc_write, clkb => clk, ceb => reg_load(r),
                                          q => dl(r));
end generate;
reg_data(7 downto 0) <= std_logic_vector(dl);

dhreg: for r in 0 to 7 generate
begin
    dhbit: component DPRegister port map (da => DBmpx(r), clka => clk, cea => lah,
                                          db => pc_write, clkb => pc_strobe, ceb => reg_load(r+8),
                                          q => dh(r));
end generate;
reg_data(15 downto 8) <= std_logic_vector(dh);

sw1reg: for r in 0 to 7 generate
signal inv_data: std_logic;
begin
    inv_data <= not datasw(r);
    sw1bit: component DPRegister port map (da => inv_data, clka => tgl_datasw(r), cea => '1',
                                           db => pc_write, clkb => pc_strobe, ceb => reg_load(r+16),
                                           q => datasw(r));
end generate;
reg_data(23 downto 16) <= datasw;

sw2reg: for r in 0 to 7 generate
begin
    toggles: if (r = 3) or (r = 0) generate    -- control switches that toggle
    signal inv_data: std_logic;
    begin
        inv_data <= not ctlsw(r);
        sw2bit: component DPRegister port map (da => inv_data, clka => tgl_ctlsw(r), cea => '1',
                                               db => pc_write, clkb => pc_strobe, ceb => reg_load(r+24),
                                               q => ctlsw(r));
    end generate;

    momentaries: if (r /= 3) and (r /= 0) generate    -- control switches that are momentary
    signal cursw: std_logic;
    begin
        sw2bit: component DPRegister port map (da => ctlsw_in(r), clka => sw2clk, cea => '1',
                                               db => pc_write, clkb => pc_strobe, ceb => reg_load(r+24),
                                               q => ctlsw(r));
    end generate;
end generate;
reg_data(31 downto 24) <= ctlsw;

-- PC interface to register file
-- the bit selected by pc_addr is always present on pc_data
-- if a pc_strobe is given, the bit is set to pc_write

process (pc_addr)   -- read registers
begin
    if pc_addr(5 downto 3) = "111" then
        pc_data <= DBcpu(to_integer(pc_addr(2 downto 0)));
    else
        pc_data <= reg_data(to_integer(pc_addr(4 downto 0)));
    end if;
end process;

-- enables for loading
process (pc_addr) -- write registers
begin
    for i in 0 to 31 loop
        if pc_addr = to_unsigned(i, pc_addr'length) then
            reg_load(i) <= '1';
         else
            reg_load(i) <= '0';
         end if;
    end loop;
end process;

-- two-phase clock generator

process (clk)
begin
    if falling_edge(clk) then
        clkdiv <= clkdiv + 1;
    end if;
end process;

ph1     <= clkdiv(2) and clkdiv(3);
ph2     <= clkdiv(2) and not clkdiv(3);
ph1CPU  <= ph1;
ph2CPU  <= ph2;

-- state decoding

st      <= (SYcpu = '0') and (ph2 = '1');
st1i    <= '1' when (Scpu = "000") and st else '0';
st1     <= '1' when (Scpu = "001") and st else '0';
st2     <= '1' when (Scpu = "010") and st else '0';
st3     <= '1' when (Scpu = "011") and st else '0';
ststop  <= '1' when (Scpu = "110") and st else '0';
stwait  <= '1' when (Scpu = "111") and st else '0';

st3_n   <= not st3;

stag_ff: FDCP port map (C => st3_n, CE => '1', PRE => st2, D => '0', CLR => pcw, Q => stag);
t3a     <= SYcpu and stag;

-- memory timing

pcc     <= cc1 and not cc2;
pcw     <= cc1 and cc2;
mem     <= t3a and not pcc;

-- RAM/ROM

CERAM_n <= not (RAMSelect and mem and not ijam);
CEROM_n <= not (mem and not RAMSelect and not ijam);
we_n    <= not (pcw or DEPwrite);

AddrB   <= RAMAddress when RAMSelect = '1' else ROMAddress;

-- interrupt timing

jammer: FDCP port map (PRE => st1i, CE => '1', CLR => st3, D => '0', C => '0', Q => ijam);

isync1: FDC port map (C => INTR, CE => '1', D => '1', CLR => st1i, Q => syncint);
isync2: FDC port map (C => ph2, CE => '1', D => syncint, CLR => st1i, Q => INTcpu);

-- condition and address registers

cc1 <= dh(6);
cc2 <= dh(7);

-- IO ports

-- are we INPuting or OUTputing data?
io      <= (t3a and pcc) = '1';

-- decode 4 input and 4 output ports "one-hot" style
portsel <= (dh(4) & dh(2) & dh(1));

io_strobe(0) <= '1' when io and (portsel = "000") else '0';
io_strobe(1) <= '1' when io and (portsel = "100") else '0';
io_strobe(2) <= '1' when io and (portsel = "001") else '0';
io_strobe(3) <= '1' when io and (portsel = "101") else '0';

-- if OUTputting, send DL to IO bus
io_pio  <= std_logic_vector(dl) when io and (dh(4) = '1') else (others => 'Z');

-- if INPuting, send UIB to CPU bus (OC output)
driveINP: for bit in 0 to 7 generate
begin
    DBcpu(bit) <= '0' when io and (dh(4) = '0') and (io_pio(bit) = '0') else 'Z';
end generate;

-- special: latch port 0 for control panel

process (t3a, io, portsel, dl)
begin
    if rising_edge(st3) then
        if io and portsel = "000" then
            port0 <= std_logic_vector(dl);
        end if;
    end if;
end process;

-- UI clock.  Divide CPU clock down so that all LEDs are serviced 15-30
-- times a second

uiclk <= scanclk(11);

process (clkdiv(3))
begin
    if rising_edge(clkdiv(3)) then
        scanclk <= scanclk + 1;
    end if;
end process;

-- UI state sequencer

process (uiclk, state)
begin
    if rising_edge(uiclk) then
        case state is

            when LEDhi => state <= LEDlo;
            when LEDlo => state <= LEDo;
            when LEDo  => state <= LEDp0;
            when LEDp0 => state <= LEDsw1;
            when LEDsw1=> state <= LEDsw2;
            when LEDsw2=> state <= SW1;
            when SW1   => state <= SW2;
            --when SW2   => state <= LEDhi;
            when others => state <= LEDhi;

        end case;
    end if;
end process;

-- strobes

LED_a <= LEDstrobes;

-- output HI address to LEDs

LED_a(0) <= '1' when state = LEDhi else '0';
LED_c <= std_logic_vector(dh) when state = LEDhi else (others => 'Z');

-- output LO address to LEDs

LED_a(1) <= '1' when state = LEDlo else '0';
LED_c <= std_logic_vector(dl) when state = LEDlo else (others => 'Z');

-- output O data to LEDs

LED_a(2) <= '1' when state = LEDo else '0';
LED_c <= DBcpu when state = LEDo else (others => 'Z');

-- output port 0 to LEDs

LED_a(3) <= '1' when state = LEDp0 else '0';
LED_c <= port0 when state = LEDp0 else (others => 'Z');

-- output data status indicators to LEDs

LED_a(4) <= '1' when state = LEDsw1 else '0';
LED_c <= datasw when state = LEDsw1 else (others => 'Z');

-- output ctrl status indicators to LEDs

LED_a(5) <= '1' when state = LEDsw2 else '0';
LED_c(3 downto 0) <= (ctlsw(3) & not ctlsw(3) & ctlsw(0) & not ctlsw(0)) when state = LEDsw2 else
                   (others => 'Z');

-- latch and debounce data switch register; each press toggles bits

sw1clk <= '1' when state = SW1 else '0';

gensw1: for sw in 0 to 7 generate
begin
    -- debounce data switches toggle switch register bits
    sw1deb: debouncer port map (clk => sw1clk, key => sw1in(sw), q => tgl_datasw(sw));
end generate;

-- latch and debounce control switches

sw2clk <= '1' when state = SW2 else '0';

gensw2: for sw in 0 to 7 generate
begin
    toggles: if (sw = 3) or (sw = 0) generate    -- control switches that toggle
    begin
        tlgctl: debouncer port map (clk => sw2clk, key => sw2in(sw), q => tgl_ctlsw(sw));
    end generate;
    
    momentaries: if (sw /= 3) and (sw /= 0) generate    -- control switches that are momentary
    begin
        momctl: debouncer port map (clk => sw2clk, key => sw2in(sw), q => ctlsw_in(sw));
    end generate;
end generate;

-- LAL, LAH switches

  -- break out switches
(lahsw, lalsw, depsw, exsw, runstepsw, stepsw, intsw, jamintsw) <= ctlsw;

lal     <= st1 or lalsw or latchdl;
lah     <= stag or lahsw or latchdh;

-- DEP/EX switches

-- trigger state machine on press of either DEP or EX

process (depsw, DEPflag2)
begin
    if rising_edge(depsw) then
        DEPflag1 <= not DEPflag2;
    end if;
end process;

process (exsw, EXflag2)
begin
    if rising_edge(exsw) then
        EXflag1 <= not EXflag2;
    end if;
end process;

-- this state machine is triggered by DEP or EX being pressed
-- its main task is to increment the address registers
-- secondarily it generates a write pulse when DEP is pressed
process (uiclk, DEPflag1, DEPflag2, EXflag1, EXflag2)
begin
    if rising_edge(uiclk) then
        case DEPEX is

            when DEStop =>
                if (DEPflag1 /= DEPflag2) or (EXflag1 /= EXflag2) then
                    DEPflag2 <= DEPflag1;   -- reset press detection
                    EXflag2  <= EXflag1;
                    DEPEX <= DELow;
                end if;

            when DELow =>
                DEPwrite <= depsw;          -- generate a write pulse if DEP pressed
                addrg <= ("0" & dl) + 1;    -- increment low part of address, saving carry
                DEPEX <= DELatchLow;

            when DELatchLow =>
                DEPwrite <= '0';            -- clear write pulse
                DEPEX <= DEHigh;            -- latch the new low part of the address

            when DEHigh =>                  -- carry into high part of address
                addrg(5 downto 0) <= dh(5 downto 0) + carry;
                DEPEX <= DELatchHigh;

            when DELatchHigh =>
                DEPEX <= DEStop;            -- latch the new high part of the address

        end case;
    end if;
end process;

-- send switch data to RAM
driveJAM: for bit in 0 to 7 generate -- drive OC
begin
    DBcpu(bit) <= '0' when (DEPEX /= DEStop) and (datasw(bit) = '0') else 'Z';
end generate;

carry   <= addrg(8 downto 8);
latchdl <= '1' when DEPEX = DELatchLow else '0';
latchdh <= '1' when DEPEX = DELatchHigh else '0';

-- RUN/STEP and STEP switches

stepff: FDC port map (C => stepsw, CE => '1', D => '1', CLR => st3, Q => step1);
RDYcpu  <= '1' when (step1 = '1') or (runstepsw = '1') else '0';

-- INT switch

intr <= intsw;

-- JAM/INT switch

DBmpx <= std_logic_vector(addrg(7 downto 0)) when (DEPEX /= DEStop) else
         datasw when jamintsw = '1' else
         DBcpu;

end RTL;

